/*	$NetBSD$	*/

/*-
 * Copyright (c) 2001 Minoura Makoto.
 * Copyright (c) 1994 Charles M. Hannum.
 * Copyright (c) 1992, 1993 Erik Forsberg.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * THIS SOFTWARE IS PROVIDED BY ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL I BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Synaptics TouchPad driver.
 * Intended to provide common functions among PS/2, serial and ADB.
 * Based on the specification available from Synaptics.
 *   (http://www.synaptics.com/support/dev_support.cfm)
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/ioctl.h>

#include <machine/bus.h>

#include <dev/ic/stpreg.h>
#include <dev/ic/stpvar.h>

#include <dev/wscons/wsconsio.h>
#include <dev/wscons/wsmousevar.h>

static char UNK[] = "Unknown sensor type";
static char *stp_idstring[] = {
	UNK,
	"TM41xx134 Standard TouchPad",
	"TM41xx156 Mini module",
	"TM41xx180 Super module",
	UNK, UNK, UNK, UNK,
	"TM41xx220 Ultra-thin module",
	"TM41xx230 Wide pad module",
	UNK,
	"TM41xx240 Stamp pad module",
	"TM41xx140 SubMini module",
	UNK, UNK,
	"TM41xx301 Advanced Technology Pad",
	"TM41xx221 Ultra-thin module, connector reserved"
};
#define STP_NID		(sizeof(stp_idstring) / sizeof(stp_idstring[0]))

static int stp_init (struct stp_softc *);
static void stp_power (int, void *);
static int stp_find_up (struct stp_softc *, int);
static int stp_get_region (struct stp_softc *, int, int);
static int stp_avg_x (struct stp_softc *, int);
static int stp_avg_y (struct stp_softc *, int);

int stp_attach (arg)
	void *arg;
{
	struct stp_softc *sc = arg;
	int r;
	stp_identifier *ident;
	struct wsmousedev_attach_args a;

	ident = &sc->sc_ident;
	r = (sc->sc_identify) (arg, ident);
	if (r) {
		printf("\nstp_init: identification failed\n");
		return EIO;
	}
	printf(": Synaptics TouchPad rev. %d.%d\n",
	       ident->si_info_major, ident->si_info_minor);

	r = ((ident->si_model_id & STP_MODELID_INFO_SENSOR_MASK)
	     >> STP_MODELID_INFO_SENSOR_SHIFT);
	printf("%s: %s", sc->sc_dev.dv_xname,
	       (r < STP_NID)?stp_idstring[r]:UNK);
	r = ((ident->si_model_id & STP_MODELID_INFO_GEOMETRY_MASK)
	     >> STP_MODELID_INFO_GEOMETRY_SHIFT);
	if (r == 1)
		printf(", rectangular");
	else if (r == 2)
		printf(", oval");
	if (ident->si_model_id & STP_MODELID_INFO_NEW_ABS)
		printf(", new packet");
	if (ident->si_model_id & STP_MODELID_CAP_PEN)
		printf(", pen capable");
	printf(".\n%s: setting absolute mode.\n", sc->sc_dev.dv_xname);
	if (stp_init(sc)) {
		printf("stp_init: setmode failed\n");
		return EIO;
	}

	/* default bindings */
	sc->sc_tap_anywhere = 1;
	sc->sc_tap_lt = 0;
	sc->sc_tap_rb = 2;
	sc->sc_tap_rt = 3;
	sc->sc_tap_lb = 6;
	sc->sc_zaxis_t = sc->sc_zaxis_b = sc->sc_zaxis_l = 0;
	sc->sc_zaxis_r = 1;

	/*
	 * Attach the wsmouse, saving a handle to it.
	 */
	a.accessops = sc->sc_wsaccess;
	a.accesscookie = arg;
	sc->sc_wsmousedev = config_found(arg, &a, wsmousedevprint);

	sc->sc_powercookie = powerhook_establish (stp_power, sc);

	return 0;
}

static int stp_init (sc)
	struct stp_softc *sc;
{
	if ((sc->sc_setmode) (sc, STP_MODE_ABSOLUTE))
		return EIO;

	sc->sc_seqno = 0;
	sc->sc_buttons = 0;
	sc->sc_status = STP_IDLE;
	sc->sc_lookat = STP_LOOK_AT;
	sc->sc_zthresh = STP_ZTHRESH_DEFAULT;

	return 0;
}

static void stp_power (why, arg)
	int why;
	void *arg;
{
	struct stp_softc *sc = arg;

	switch(why) {
	case PWR_RESUME:
		stp_init (sc);
		break;
	default:
		break;
	}
}

static int stp_find_up (sc, lookat)
	struct stp_softc *sc;
	int lookat;
{
	int i;

	for (i = lookat-1; i >= 0; i--) {
		if (sc->sc_history[i].sa_z < sc->sc_zthresh)
			return i;
	}

	return -1;
}

static int stp_get_region (sc, down, up)
	struct stp_softc *sc;
	int down, up;
{
	int x = 0, y = 0;
	int i, r = 0;

	for (i = down; i > up; i--) {
		x += sc->sc_history[i].sa_x;
		y += sc->sc_history[i].sa_y;
	}
	x /= down - up;
	y /= down - up;

	if (x < STP_COORDINATE_X_EDGE_MIN)
		r |= LEFT;
	else if (x > STP_COORDINATE_X_EDGE_MAX)
		r |= RIGHT;
	if (y < STP_COORDINATE_Y_EDGE_MIN)
		r |= TOP;
	else if (y > STP_COORDINATE_Y_EDGE_MAX)
		r |= BOTTOM;

	return r;
}

static int stp_avg_x (sc, h)
	struct stp_softc *sc;
	int h;
{
	int i, n, x;

	for (i = n = x = 0; i < STP_AVG_NSAMPLES; i++) {
		if (sc->sc_history[h + i].sa_x > 0) {
			x += sc->sc_history[h + i].sa_x;
			n++;
		}
	}
	x /= n;

	return x;
}

static int stp_avg_y (sc, h)
	struct stp_softc *sc;
	int h;
{
	int i, n, y;

	for (i = n = y = 0; i < STP_AVG_NSAMPLES; i++) {
		if (sc->sc_history[h + i].sa_y > 0) {
			y += sc->sc_history[h + i].sa_y;
			n++;
		}
	}
	y /= n;

	return y;
}

void stp_input (arg, p)
	void *arg;
	struct stp_abspacket *p;
{
	struct stp_softc *sc = arg;
	int buttons = 0;
	int foundinput = 0;
	int x, y, dx = 0, dy = 0, dz = 0;

	memmove(&sc->sc_history[1], &sc->sc_history[0],
		sizeof (sc->sc_history) - sizeof (sc->sc_history[0]));
	sc->sc_history[0] = *p;

	if (sc->sc_status == STP_WAIT) {
		if (++sc->sc_lookat < STP_LOOK_AT){
			return;}
		sc->sc_status = STP_IDLE;
	}

	/* Process (physical) buttons */
	if (sc->sc_history[sc->sc_lookat].sa_lbutton)
		buttons |= 1;
	if (sc->sc_history[sc->sc_lookat].sa_rbutton)
		buttons |= 4;
	if (sc->sc_buttons ^ buttons) {
		sc->sc_buttons = buttons;
		foundinput++;
	}

	/* Process movement */
	switch (sc->sc_status) {
	case STP_IDLE:
		if (sc->sc_history[sc->sc_lookat].sa_z >= sc->sc_zthresh) {
			int region;
			int toff;

			toff = stp_find_up(sc, sc->sc_lookat);
			if (toff >= 0 && sc->sc_buttons == 0 &&
			    sc->sc_lookat - toff <= STP_TAP_MAX_DURATION) {
				sc->sc_status = STP_TAPPING;
				region = stp_get_region(sc,
							sc->sc_lookat, toff);
				if (sc->sc_tap_lt &&
				    region == (LEFT|TOP))
					buttons |= 1 << (sc->sc_tap_lt - 1);
				else if (sc->sc_tap_rt &&
					 region == (RIGHT|TOP))
					buttons |= 1 << (sc->sc_tap_rt - 1);
				else if (sc->sc_tap_lb &&
					 region == (LEFT|BOTTOM))
					buttons |= 1 << (sc->sc_tap_lb - 1);
				else if (sc->sc_tap_rb &&
					 region == (RIGHT|BOTTOM))
					buttons |= 1 << (sc->sc_tap_rb - 1);
				else
					buttons |= 1<<(sc->sc_tap_anywhere-1);
				foundinput++;
			} else {
				region = stp_get_region(sc, 
					    sc->sc_lookat, sc->sc_lookat-1);
				if ((sc->sc_zaxis_t && (region & TOP)) ||
				    (sc->sc_zaxis_b && (region & BOTTOM))) {
					sc->sc_scraxis = 1;
					sc->sc_scrz = 0;
					sc->sc_status = STP_SCROLLING;
				} else if ((sc->sc_zaxis_l && (region&LEFT)) ||
				    (sc->sc_zaxis_r && (region & RIGHT))) {
					sc->sc_scraxis = 0;
					sc->sc_scrz = 0;
					sc->sc_status = STP_SCROLLING;
				} else {
					sc->sc_status = STP_XYMOVE;
				}
				sc->sc_x = sc->sc_history[sc->sc_lookat].sa_x;
				sc->sc_y = sc->sc_history[sc->sc_lookat].sa_y;
			}
		}
		break;

	case STP_TAPPING:
		if (sc->sc_history[sc->sc_lookat].sa_z < sc->sc_zthresh) {
			/* button up */
			foundinput++;
			buttons = sc->sc_buttons;
			sc->sc_status = STP_IDLE;
		}
		break;

	case STP_XYMOVE:
		while (1) {
			int dx0, dy0;

			x = stp_avg_x (sc, sc->sc_lookat);
			y = stp_avg_y (sc, sc->sc_lookat);
			if (x) {
				dx0 = x - sc->sc_x;
				if (dx0 >= 0)
					dx += dx0*dx0 / 1500;
				else
					dx -= dx0*dx0 / 1500;
			}
			if (y) {
				dy0 = y - sc->sc_y;
				if (dy0 >= 0)
					dy = dy0*dy0 / 1000;
				else
					dy -= dy0*dy0 / 1000;
			}
			if (sc->sc_lookat == 0)
				break;
			sc->sc_lookat--;
		}
		if (dx > 127)
			dx = 127;
		if (dx < -127)
			dx = -127;
		if (dy > 127)
			dy = 127;
		if (dy < -127)
			dy = -127;
		if (dx || dy) {
			foundinput++;
		}
		if (sc->sc_history[sc->sc_lookat].sa_z < sc->sc_zthresh) {
			sc->sc_status = STP_WAIT;
		} else {
			sc->sc_x = x;
			sc->sc_y = y;
		}
		break;

	case STP_SCROLLING:
		if (sc->sc_scraxis) {
			x = stp_avg_x (sc, sc->sc_lookat);
			dz = sc->sc_x - x;
			sc->sc_x = x;
		} else {
			y = stp_avg_y (sc, sc->sc_lookat);
			dz = sc->sc_y - y;
			sc->sc_y = y;
		}
		sc->sc_scrz += dz;
		if (sc->sc_scrz > 300) {	/* XXX: constant */
			sc->sc_scrz -= 300;
			foundinput++;
			dz = 1;
		} else if (sc->sc_scrz < -300) {	/* XXX: constant */
			sc->sc_scrz += 300;
			foundinput++;
			dz = -1;
		}
		if (sc->sc_history[sc->sc_lookat].sa_z < sc->sc_zthresh)
			sc->sc_status = STP_IDLE;
		break;

	case STP_DRAGGING:
	case STP_WAIT:
		break;
	}

	if (foundinput)
		wsmouse_input(sc->sc_wsmousedev,
			      buttons, dx, dy, dz,
			      WSMOUSE_INPUT_DELTA);
}
